#include "archive.h"
#include "extapp_api.h"
#include "../global_preferences.h"

#include <string.h>
#include <stdlib.h>

static char print_hex(int val){
  val &= 0xf;
  if (val>=0 && val<=9)
    return '0'+val;
  return 'A'+(val-10);
}

static void print_hex(unsigned val,char * ptr){
  for (int i=0;i<8;++i){
    ptr[i]=print_hex(val>>(28-4*i));
  }
}

/*********************************************************************
   * Filename:   sha256.c/.h
   * Author:     Brad Conte (brad AT bradconte.com)
   * Copyright:
   * Disclaimer: This code is presented "as is" without any guarantees.
   * Details:    Implementation of the SHA-256 hashing algorithm.
              SHA-256 is one of the three algorithms in the SHA2
              specification. The others, SHA-384 and SHA-512, are not
              offered in this implementation.
              Algorithm specification can be found here:
	      * http://csrc.nist.gov/publications/fips/fips180-2/fips180-2withchangenotice.pdf
              This implementation uses little endian byte order.
  *********************************************************************/
#define SHA256_BLOCK_SIZE 32            // SHA256 outputs a 32 byte digest
  
  /**************************** DATA TYPES ****************************/
  typedef unsigned char BYTE;             // 8-bit byte
  typedef unsigned int  WORD;             // 32-bit word, change to "long" for 16-bit machines
  
  typedef struct {
    BYTE data[64];
    WORD datalen;
    unsigned long long bitlen;
    WORD state[8];
  } SHA256_CTX;
  
  
  /****************************** MACROS ******************************/
#define ROTLEFT(a,b) (((a) << (b)) | ((a) >> (32-(b))))
#define ROTRIGHT(a,b) (((a) >> (b)) | ((a) << (32-(b))))

#define CH(x,y,z) (((x) & (y)) ^ (~(x) & (z)))
#define MAJ(x,y,z) (((x) & (y)) ^ ((x) & (z)) ^ ((y) & (z)))
#define EP0(x) (ROTRIGHT(x,2) ^ ROTRIGHT(x,13) ^ ROTRIGHT(x,22))
#define EP1(x) (ROTRIGHT(x,6) ^ ROTRIGHT(x,11) ^ ROTRIGHT(x,25))
#define SIG0(x) (ROTRIGHT(x,7) ^ ROTRIGHT(x,18) ^ ((x) >> 3))
#define SIG1(x) (ROTRIGHT(x,17) ^ ROTRIGHT(x,19) ^ ((x) >> 10))

  /**************************** VARIABLES *****************************/
  static const WORD k[64] = {
    0x428a2f98,0x71374491,0xb5c0fbcf,0xe9b5dba5,0x3956c25b,0x59f111f1,0x923f82a4,0xab1c5ed5,
    0xd807aa98,0x12835b01,0x243185be,0x550c7dc3,0x72be5d74,0x80deb1fe,0x9bdc06a7,0xc19bf174,
    0xe49b69c1,0xefbe4786,0x0fc19dc6,0x240ca1cc,0x2de92c6f,0x4a7484aa,0x5cb0a9dc,0x76f988da,
    0x983e5152,0xa831c66d,0xb00327c8,0xbf597fc7,0xc6e00bf3,0xd5a79147,0x06ca6351,0x14292967,
    0x27b70a85,0x2e1b2138,0x4d2c6dfc,0x53380d13,0x650a7354,0x766a0abb,0x81c2c92e,0x92722c85,
    0xa2bfe8a1,0xa81a664b,0xc24b8b70,0xc76c51a3,0xd192e819,0xd6990624,0xf40e3585,0x106aa070,
    0x19a4c116,0x1e376c08,0x2748774c,0x34b0bcb5,0x391c0cb3,0x4ed8aa4a,0x5b9cca4f,0x682e6ff3,
    0x748f82ee,0x78a5636f,0x84c87814,0x8cc70208,0x90befffa,0xa4506ceb,0xbef9a3f7,0xc67178f2
  };

  /*********************** FUNCTION DEFINITIONS ***********************/
  void giac_sha256_transform(SHA256_CTX *ctx, const BYTE data[])
  {
    WORD a, b, c, d, e, f, g, h, i, j, t1, t2, m[64];

    for (i = 0, j = 0; i < 16; ++i, j += 4)
      m[i] = (data[j] << 24) | (data[j + 1] << 16) | (data[j + 2] << 8) | (data[j + 3]);
    for ( ; i < 64; ++i)
      m[i] = SIG1(m[i - 2]) + m[i - 7] + SIG0(m[i - 15]) + m[i - 16];

    a = ctx->state[0];
    b = ctx->state[1];
    c = ctx->state[2];
    d = ctx->state[3];
    e = ctx->state[4];
    f = ctx->state[5];
    g = ctx->state[6];
    h = ctx->state[7];

    for (i = 0; i < 64; ++i) {
      t1 = h + EP1(e) + CH(e,f,g) + k[i] + m[i];
      t2 = EP0(a) + MAJ(a,b,c);
      h = g;
      g = f;
      f = e;
      e = d + t1;
      d = c;
      c = b;
      b = a;
      a = t1 + t2;
    }

    ctx->state[0] += a;
    ctx->state[1] += b;
    ctx->state[2] += c;
    ctx->state[3] += d;
    ctx->state[4] += e;
    ctx->state[5] += f;
    ctx->state[6] += g;
    ctx->state[7] += h;
  }

  void giac_sha256_init(SHA256_CTX *ctx)
  {
    ctx->datalen = 0;
    ctx->bitlen = 0;
    ctx->state[0] = 0x6a09e667;
    ctx->state[1] = 0xbb67ae85;
    ctx->state[2] = 0x3c6ef372;
    ctx->state[3] = 0xa54ff53a;
    ctx->state[4] = 0x510e527f;
    ctx->state[5] = 0x9b05688c;
    ctx->state[6] = 0x1f83d9ab;
    ctx->state[7] = 0x5be0cd19;
  }

  void giac_sha256_update(SHA256_CTX *ctx, const BYTE data[], size_t len)
  {
    WORD i;

    for (i = 0; i < len; ++i) {
      ctx->data[ctx->datalen] = data[i];
      ctx->datalen++;
      if (ctx->datalen == 64) {
	giac_sha256_transform(ctx, ctx->data);
	ctx->bitlen += 512;
	ctx->datalen = 0;
      }
    }
  }

  void giac_sha256_final(SHA256_CTX *ctx, BYTE hash[])
  {
    WORD i;

    i = ctx->datalen;

    // Pad whatever data is left in the buffer.
    if (ctx->datalen < 56) {
      ctx->data[i++] = 0x80;
      while (i < 56)
	ctx->data[i++] = 0x00;
    }
    else {
      ctx->data[i++] = 0x80;
      while (i < 64)
	ctx->data[i++] = 0x00;
      giac_sha256_transform(ctx, ctx->data);
      memset(ctx->data, 0, 56);
    }

    // Append to the padding the total message's length in bits and transform.
    ctx->bitlen += ctx->datalen * 8;
    ctx->data[63] = ctx->bitlen;
    ctx->data[62] = ctx->bitlen >> 8;
    ctx->data[61] = ctx->bitlen >> 16;
    ctx->data[60] = ctx->bitlen >> 24;
    ctx->data[59] = ctx->bitlen >> 32;
    ctx->data[58] = ctx->bitlen >> 40;
    ctx->data[57] = ctx->bitlen >> 48;
    ctx->data[56] = ctx->bitlen >> 56;
    giac_sha256_transform(ctx, ctx->data);

    // Since this implementation uses little endian byte ordering and SHA uses big endian,
    // reverse all the bytes when copying the final state to the output hash.
    for (i = 0; i < 4; ++i) {
      hash[i]      = (ctx->state[0] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 4]  = (ctx->state[1] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 8]  = (ctx->state[2] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 12] = (ctx->state[3] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 16] = (ctx->state[4] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 20] = (ctx->state[5] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 24] = (ctx->state[6] >> (24 - i * 8)) & 0x000000ff;
      hash[i + 28] = (ctx->state[7] >> (24 - i * 8)) & 0x000000ff;
    }
  }
  /* END OF SHA256 */

struct sha256_t {
  int len;
  BYTE buf[SHA256_BLOCK_SIZE];
};

#if 0
sha256_t shakeys[]={
		    {5160176, {86,64,84,218,208,31,176,238,15,35,25,99,139,26,173,220,67,173,150,124,31,135,121,30,42,70,171,19,73,152,172,250,}},
};
#else
// run app_sig in khiext to create apps_sha256.h
#include "apps_sha256.h"
#endif
    
bool sha256_check(const unsigned char * ptr,int len){
  BYTE buf[SHA256_BLOCK_SIZE];
  SHA256_CTX ctx;
  giac_sha256_init(&ctx);
  giac_sha256_update(&ctx, ptr, len);
  giac_sha256_final(&ctx, buf);
  for (unsigned i=0;i<sizeof(shakeys)/sizeof(sha256_t);++i){
    if (0){
      extapp_pushRectUniform(0,100,320,100,0);
      char tab[9]={0};
      print_hex(i,tab);
      extapp_drawTextSmall(tab,0,100,65535,0,false);
      print_hex(len,tab);
      extapp_drawTextSmall(tab,0,120,65535,0,false);
      print_hex(shakeys[i].len,tab);
      extapp_drawTextSmall(tab,0,140,65535,0,false);
      extapp_getKey(false,0);
    }
    if (len!=shakeys[i].len)
      continue;
    // now try for all compatible sizes
    // we do not know the firmware size, we extract 2M or 6M
    // the part after the end is not known
    if (!memcmp(shakeys[i].buf, buf, SHA256_BLOCK_SIZE))
      return true;
  }
  return false;
}
  

namespace External {
  namespace Archive {

#ifdef DEVICE

    static_assert(sizeof(TarHeader) == 512);

    bool isSane(const TarHeader* tar) {
      unsigned char ch=tar->name[0];
      return !memcmp(tar->magic, "ustar  ", 8) &&  ch>=32 && ch<128;
    }

    bool isExamModeAndFileNotExecutable(const TarHeader* tar) {
      return GlobalPreferences::sharedGlobalPreferences()->isInExamMode() && (tar->mode[4] & 0x01) == 0;
    }

    const int last_certified_index=8; // files 0 to 8 are certified

    bool ck_exammode(const TarHeader* tar,int index) {
      return GlobalPreferences::sharedGlobalPreferences()->isInExamMode() && index>last_certified_index;
    }

    int tar_size( const TarHeader* tar){
      if (!isSane(tar)) 
	return -1;
      size_t size = 0;
      for (int i=0; i<11; i++){
	char ch=tar->size[i];
	if (ch==' ')
	  return size;
	if (ch<'0' || ch>'7')
	  return -1;
	size = size*8 + (ch-'0');
      }
      return size;
    }

    void tar_get_entry(const TarHeader* tar,size_t size,File & entry){
      entry.name = tar->name;
      entry.data = reinterpret_cast<const uint8_t*>(tar) + sizeof(TarHeader);
      entry.dataLength = size;
      entry.isExecutable = (tar->mode[4] & 0x01)==1;
      entry.readable = (tar->mode[4] & 0x04)==1;
    }

    const TarHeader* tar_nxt(const TarHeader* tar,size_t size){
      unsigned stride = (sizeof(TarHeader) + size + 511);
      stride = (stride >> 9) << 9;
      tar = reinterpret_cast<const TarHeader*>(reinterpret_cast<const char*>(tar) + stride);
      return tar;
    }    

    bool fileAtIndex(size_t index, File &entry) {
      if (index==(size_t) -1)
	return false;  
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      // TAR file is a set of records aligned to 512 bytes boundary followed by data.
      for (size_t idx=0;(size_t) tar<0x90800000;++idx) {
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return false;
	// Check if we found our file.
	if (index==idx) {
	  // If yes, check for sanity and for exam mode stuff
	  if (!isSane(tar) || (tar->mode[4] & 0x04)==0 ||
	      ck_exammode(tar,index)
	      // isExamModeAndFileNotExecutable(tar)
	      )
	    return false;
	  // File entry found, copy data out.
	  tar_get_entry(tar,size,entry);
	  return true;
	}
	// move to the next TAR header.
	tar=tar_nxt(tar,size);
      }
  
      // Achievement unlock: How did we get there ?
      return false;
    }

    extern "C" void (* const apiPointers[])(void);
    typedef uint32_t (*entrypoint)(const uint32_t, const void *, void *, const uint32_t);

    uint32_t executeFile(const char *name, void * heap, const uint32_t heapSize) {
      File entry;
      int i=indexFromName(name);
      if(fileAtIndex(i, entry)) {
	if(!entry.isExecutable || i>last_certified_index) {
	  return 0;
	}
#ifdef DFU_PROTECT
	if (0){
	  extapp_pushRectUniform(0,100,320,100,0);
	  char tab[9]={0};
	  print_hex((unsigned)entry.data+entry.dataLength,tab);
	  extapp_drawTextSmall(tab,0,100,65535,0,false);
	  print_hex(DFU_PROTECT,tab);
	  extapp_drawTextSmall(tab,0,120,65535,0,false);
	  extapp_getKey(false,0);
	}
	// if DFU had the opportunity to overwrite (part of) the binary
	// it might be compromised, do not run it unless it's SHA256 hash
	// can be checked 
	// sha256 computation takes too much time for KhiCAS
	// that's why we raise DFU_PROTECT value after the end of KhiCAS
	if ((unsigned) entry.data+entry.dataLength>=DFU_PROTECT){
	  if (!sha256_check(entry.data,entry.dataLength)){
	    char tab[9]={0};
	    extapp_pushRectUniform(0,100,320,100,0);
	    extapp_drawTextSmall("sha256 check failure, start/length",0,100,65535,0,false);
	    print_hex((unsigned)entry.data,tab);
	    extapp_drawTextSmall(tab,0,100,65535,0,false);
	    print_hex(entry.dataLength,tab);
	    extapp_drawTextSmall(tab,0,100,65535,0,false);
	    extapp_getKey(false,0);
	    return 0;
	  }
	  else {
	    extapp_drawTextSmall("sha256 check OK",0,120,65535,0,false);
	    // extapp_getKey(false,0);
	  }
	}
#endif
	uint32_t ep = *reinterpret_cast<const uint32_t*>(entry.data);
	if(ep >= 0x90200000 && ep < 0x90800000) {
	  return ((entrypoint)ep)(API_VERSION, apiPointers, heap, heapSize);
	}
      }
      return -1;
    }

#define EFFICIENT 1

    int indexFromName(const char *name) {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      for (int idx=0;(size_t) tar<0x90800000U;++idx){
	if (!isSane(tar) ||
	    ck_exammode(tar,idx)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return -1;
	// Calculate the size
	int size = tar_size(tar);
	if (size<0)
	  return -1;
	if ( (tar->mode[4] & 0x04) && strcmp(name,tar->name)==0 )
	  return idx; // found
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return -1;
#else // old inefficient code in O(n^2)
      File entry;

      for (int i = 0; fileAtIndex(i, entry); i++) {
	if (entry.readable && strcmp(name, entry.name) == 0) {
	  return i;
	}
      }

      return -1;
#endif
    }

    size_t numberOfFiles() {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      int idx;
      for (idx=0;(size_t) tar<0x90800000U;++idx){
	if (!isSane(tar) ||
	    ck_exammode(tar,idx)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return idx;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return idx;
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return idx;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;

      for (count = 0; fileAtIndex(count, dummy); count++);

      return count;
#endif
    }

    bool executableAtIndex(size_t index, File &entry) {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000);
      size_t idx=0;
      for (int count=0;(size_t) tar<0x90800000U;++count){
	if (!isSane(tar) ||
	    ck_exammode(tar,count)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return false;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return false;
	if (tar->mode[4] & 0x01){
	  if (idx==index){
	    tar_get_entry(tar,size,entry);
	    return true;
	  }
	  idx++;
	}
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return false;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;
      size_t final_count = 0;
      // inefficient but it's not important here since exe will be at tar begin 
      for (count = 0; fileAtIndex(count, dummy); count++) {
	if (dummy.isExecutable) {
	  if (final_count == index) {
	    entry.name = dummy.name;
	    entry.data = dummy.data;
	    entry.dataLength = dummy.dataLength;
	    entry.isExecutable = dummy.isExecutable;
	    entry.readable = dummy.readable;
	    return true;
	  }
	  final_count++;
	}
      }
  
      return false;
#endif
    }

    size_t numberOfExecutables() {
#ifdef EFFICIENT
      const TarHeader* tar = reinterpret_cast<const TarHeader*>(0x90200000U);
      int idx=0;
      for (int count=0;(size_t) tar<0x90800000U;++count){
	if (!isSane(tar) ||
	    ck_exammode(tar,count)
	    // isExamModeAndFileNotExecutable(tar)
	    ) 
	  return idx;
	// Calculate the size
	int size=tar_size(tar);
	if (size<0)
	  return idx;
	if (tar->mode[4] & 0x01)
	  idx++;
	// go to next header
	tar=tar_nxt(tar,size);
      }
      return idx;
#else // old inefficient code in O(n^2)
      File dummy;
      size_t count;
      size_t final_count = 0;
      // inefficient code but not that important since all exe are at tar begin
      for (count = 0; fileAtIndex(count, dummy); count++){
	if (dummy.isExecutable)
	  final_count++;
	// else break; // all exe must be at tar begin
      }

      return final_count;
#endif
    }



#else // DEVICE

    bool fileAtIndex(size_t index, File &entry) {
      if (index != 0)
	return false;
  
      entry.name = "Built-in";
      entry.data = NULL;
      entry.dataLength = 0;
      entry.isExecutable = true;
      entry.readable = true;
      return true;
    }

    bool executableAtIndex(size_t index, File &entry) {
      return fileAtIndex(index, entry);
    }

    size_t numberOfExecutables() {
      return 1;
    }

    extern "C" void extapp_main(void);

    uint32_t executeFile(const char *name, void * heap, const uint32_t heapSize) {
      extapp_main();
      return 0;
    }

    int indexFromName(const char *name) {
      if (strcmp(name, "Built-in") == 0)
	return 0;
      else
	return -1;
    }

    size_t numberOfFiles() {
      return 1;
    }

#endif // DEVICE

  }
}
